## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
# .libPaths(.Library)

## ----load libraries-----------------------------------------------------------
suppressPackageStartupMessages({
  library(anglemania)
  library(dplyr)
  library(Seurat)
  library(splatter)
  library(SingleCellExperiment)
  library(scater)
  library(scran)
  library(bluster)
  library(batchelor)
})

## ----create simulated data----------------------------------------------------
batch.facLoc <- 0.1
de.facLoc <- 0.1
nBatches <- 4
nGroups <- 3
nGenes <- 5000
groupCells <- 300

sce <- splatSimulate(
    batchCells = rep(groupCells * nGroups, nBatches),
    batch.facLoc = batch.facLoc,
    group.prob = rep(1/nGroups, nGroups),
    nGenes = nGenes,
    batch.facScale = 0.1,
    method = "groups",
    verbose = FALSE,
    out.prob    = 0.001,
    de.prob     = 0.1, # mild
    de.facLoc   = de.facLoc,
    de.facScale = 0.1,
    bcv.common  = 0.1,
    seed = 42
)
sce

## -----------------------------------------------------------------------------
sce_unintegrated <- sce
# Normalization.
sce_unintegrated <- logNormCounts(sce_unintegrated)

# Feature selection.
dec <- modelGeneVar(sce_unintegrated)
hvg <- getTopHVGs(dec, prop = 0.1)

# PCA.
set.seed(1234)
sce_unintegrated <- scater::runPCA(sce_unintegrated, ncomponents = 50, subset_row = hvg)

# Clustering.
colLabels(sce_unintegrated) <- clusterCells(sce_unintegrated,
  use.dimred = "PCA",
  BLUSPARAM = NNGraphParam(cluster.fun = "louvain")
)

# Visualization.
sce_unintegrated <- scater::runUMAP(sce_unintegrated, dimred = "PCA")
plotUMAP(sce_unintegrated, colour_by = "Batch")
plotUMAP(sce_unintegrated, colour_by = "Group")

## ----create anglemania object-------------------------------------------------
head(colData(sce))
batch_key <- "Batch"
angl <- create_anglemania_object(
  sce,
  batch_key = batch_key
)
angl

## ----run anglemania-----------------------------------------------------------
angl <- anglemania(angl,
  zscore_mean_threshold = 2,
  zscore_sn_threshold = 2,
  max_n_genes = 2000 # optionally define a max number of genes.
)
# If you think the number of selected genes is too high or low you can adjust the thresholds:
angl <- select_genes(angl,
                     zscore_mean_threshold = 2.5,
                     zscore_sn_threshold = 2.5)
# Inspect the anglemania genes
integration_genes <- get_anglemania_genes(angl)
head(integration_genes)
length(integration_genes)

## ----echo=FALSE, results="hide"-----------------------------------------------
hvg <- sce %>%
  scater::logNormCounts() %>%
  modelGeneVar(block = colData(sce)[[batch_key]]) %>%
  getTopHVGs(n = 320)
  
barcodes_by_batch <- split(rownames(colData(sce)), colData(sce)[[batch_key]])
sce_list <- lapply(barcodes_by_batch, function(x) sce[, x])
sce_mnn <- sce %>%
  scater::logNormCounts()
sce_mnn <- batchelor::fastMNN(sce_mnn, subset.row = hvg, k = 20, batch = factor(colData(sce_mnn)[[batch_key]]), d = 50)
reducedDim(sce, "MNN_hvg_320") <- reducedDim(sce_mnn, "corrected")
sce <- scater::runUMAP(sce, dimred = "MNN_hvg_320", name = "umap_MNN_hvg_320")
# k is the number of nearest neighbours to consider when identifying MNNs

## ----echo=FALSE, results="hide"-----------------------------------------------
hvg <- sce %>%
  scater::logNormCounts() %>%
  modelGeneVar(block = colData(sce)[[batch_key]]) %>%
  getTopHVGs(n = 2000)
  
barcodes_by_batch <- split(rownames(colData(sce)), colData(sce)[[batch_key]])
sce_list <- lapply(barcodes_by_batch, function(x) sce[, x])
sce_mnn <- sce %>%
  scater::logNormCounts()
sce_mnn <- batchelor::fastMNN(sce_mnn, subset.row = hvg, k = 20, batch = factor(colData(sce_mnn)[[batch_key]]), d = 50)
reducedDim(sce, "MNN_hvg_2000") <- reducedDim(sce_mnn, "corrected")
sce <- scater::runUMAP(sce, dimred = "MNN_hvg_2000", name = "umap_MNN_hvg_2000")

## ----echo=FALSE, results="hide"-----------------------------------------------
sce_mnn <- sce %>%
  scater::logNormCounts()
sce_mnn <- batchelor::fastMNN(sce_mnn, subset.row = integration_genes, k = 20, batch = factor(colData(sce_mnn)[[batch_key]]), d = 50)
reducedDim(sce, "MNN_anglemania") <- reducedDim(sce_mnn, "corrected")
sce <- scater::runUMAP(sce, dimred = "MNN_anglemania", name = "umap_MNN_anglemania")

## -----------------------------------------------------------------------------
plotReducedDim(sce, colour_by = "Batch", dimred = "umap_MNN_hvg_320") +
  ggtitle("MNN integration using top 320 HVGs, colored by Batch")
plotReducedDim(sce, colour_by = "Group", dimred = "umap_MNN_hvg_320") +
  ggtitle("MNN integration using top 320 HVGs, colored by Group")
plotReducedDim(sce, colour_by = "Batch", dimred = "umap_MNN_hvg_2000") +
  ggtitle("MNN integration using top 2000 HVGs, colored by Batch")
plotReducedDim(sce, colour_by = "Group", dimred = "umap_MNN_hvg_2000") +
  ggtitle("MNN integration using top 2000 HVGs, colored by Group")
plotReducedDim(sce, colour_by = "Batch", dimred = "umap_MNN_anglemania") +
  ggtitle("MNN integration using anglemania genes, colored by Batch")
plotReducedDim(sce, colour_by = "Group", dimred = "umap_MNN_anglemania") +
  ggtitle("MNN integration using anglemania genes, colored by Group")

## -----------------------------------------------------------------------------
sessionInfo()

