## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
# .libPaths(.Library)

## ----load libraries-----------------------------------------------------------
suppressPackageStartupMessages({
  library(anglemania)
  library(dplyr)
  library(Seurat)
  library(splatter)
})

## ----create simulated data----------------------------------------------------
batch.facLoc <- 0.3
de.facLoc <- 0.1
nBatches <- 4
nGroups <- 3
nGenes <- 5000
groupCells <- 300

sim <- splatSimulate(
    batchCells = rep(groupCells * nGroups, nBatches),
    batch.facLoc = batch.facLoc,
    group.prob = rep(1/nGroups, nGroups),
    nGenes = nGenes,
    batch.facScale = 0.1,
    method = "groups",
    verbose = FALSE,
    out.prob    = 0.001,
    de.prob     = 0.1, # mild 
    de.facLoc   = de.facLoc,
    de.facScale = 0.1,
    bcv.common  = 0.1,
    seed = 42
)

se <- CreateSeuratObject(counts = counts(sim), meta.data = as.data.frame(colData(sim)))
rm(sim)
se
head(se[[]])

## -----------------------------------------------------------------------------

se_unintegrated <- se
suppressWarnings({
  se_unintegrated <- se_unintegrated %>%
  NormalizeData() %>%
  FindVariableFeatures() %>%
  ScaleData() %>%
  RunPCA() %>%
  RunUMAP(dims = 1:30, verbose = FALSE)
})
DimPlot(se_unintegrated, reduction = "umap", group.by = "Batch")
DimPlot(se_unintegrated, reduction = "umap", group.by = "Group")

## ----create anglem object-----------------------------------------------------
head(se[[]])
batch_key <- "Batch"

angl <- create_anglemania_object(se,
  batch_key = batch_key,
  min_cells_per_gene = 1
)
angl

## ----run anglemania-----------------------------------------------------------
angl <- anglemania(angl,
  zscore_mean_threshold = 2,
  zscore_sn_threshold = 2,
  max_n_genes = 2000 # optionally define a max number of genes.
)
angl <- select_genes(angl,
                     zscore_mean_threshold = 2.5,
                     zscore_sn_threshold = 2.5)
# Inspect the anglemania genes
integration_genes <- get_anglemania_genes(angl)
head(integration_genes)
length(integration_genes)

## ----integration option 1-----------------------------------------------------
# if the Seurat FindIntegrationAnchors() function does not work,
# change this to the specified size:
options(future.globals.maxSize = 4000 * 1024^2)
suppressWarnings({
  seurat_integrated_angl <- integrate_by_features(se,
    angl,
    process = TRUE
  )
})
seurat_integrated_angl

## -----------------------------------------------------------------------------
Seurat::DimPlot(seurat_integrated_angl, reduction = "umap", group.by = "Batch")
Seurat::DimPlot(seurat_integrated_angl, reduction = "umap", group.by = "Group")

## ----get HVGs-----------------------------------------------------------------
se_list <- SplitObject(se, split.by = batch_key)
se_list <- lapply(se_list, NormalizeData)
# anglemania chose only 319 genes based on the thresholds
#   to make it comparable we only choose 320 highly variable genes:
hvg_features <- Seurat::SelectIntegrationFeatures(se_list, nfeatures = 320) 

## ----integration option 2-----------------------------------------------------
suppressWarnings({
  seurat_integrated_hvg <- integrate_seurat_list(se_list,
    features = hvg_features,
    process = TRUE
  )
})
seurat_integrated_hvg

## -----------------------------------------------------------------------------
Seurat::DimPlot(seurat_integrated_hvg, reduction = "umap", group.by = "Batch")
Seurat::DimPlot(seurat_integrated_hvg, reduction = "umap", group.by = "Group")

## -----------------------------------------------------------------------------
sessionInfo()

